import aio_pika
import asyncio
import inspect
import logging
import pydantic
import socket
import time

from . import errors
from .models import Request, Response
from .types import Message, Priority
from .utils import fix_message_timestamp


logger = logging.getLogger(__name__)


class RpcServer:
    def __init__(self, service_name):
        self._service_name = service_name
        self._app_id = f"{service_name}@{socket.gethostname()}"

        self._handlers = {}

        self._reset()

        self.bind(self._ping)
        self.bind(self._help)

    def _reset(self):
        self._connection = None
        self._channel = None
        self._exchange = None
        self._queue = None

    def bind(self, handler):
        if handler.__name__ in self._handlers:
            raise KeyError(f"Handler {handler.__name__} already defined")

        self._validate_handler(handler)

        self._handlers[handler.__name__] = pydantic.validate_call(
            handler,
            validate_return=True,
        )

        return handler

    def _validate_handler(self, handler):
        if not inspect.iscoroutinefunction(handler):
            raise TypeError("Handlers must be coroutine functions")

        positional_only = []
        positional_or_keyword = []

        for parameter_name, parameter in inspect.signature(handler).parameters.items():
            if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                positional_or_keyword.append(parameter_name)

            elif parameter.kind in [
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.VAR_POSITIONAL,
            ]:
                positional_only.append(parameter_name)

        if positional_only:
            raise TypeError(
                "{} has positional-only parameters {} that are not supported in {}".format(
                    handler,
                    positional_only,
                    self.__class__,
                )
            )

        if positional_or_keyword:
            logger.warning(
                "%s has positional parameters %s, only keyword parameters are supported in %s",
                handler,
                positional_or_keyword,
                self.__class__,
            )

    async def start(self, connection_string, prefetch_count=1):
        if self._connection:
            raise RuntimeError("Already started")

        connection = self._connection = await aio_pika.connect_robust(connection_string)
        channel = self._channel = await connection.channel()
        await channel.set_qos(prefetch_count=prefetch_count)

        exchange = self._exchange = await channel.declare_exchange(
            name=f"service.{self._service_name}",
            type=aio_pika.ExchangeType.TOPIC,
            durable=True,
        )

        queue = self._queue = await channel.declare_queue(
            name=f"service.{self._service_name}",
            durable=True,
            arguments={
                "x-max-priority": max(Priority),
            },
        )

        await queue.bind(exchange, "#")
        await queue.consume(fix_message_timestamp(self._on_message))

        logger.info("RPC server is consuming messages from `%s`", queue)

    async def stop(self):
        connection = self._connection
        if connection:
            self._reset()
            await connection.close()

    async def serve_forever(self, connection_string):
        await self.start(connection_string)
        try:
            await asyncio.Future()
        finally:
            await self.stop()

    async def _on_message(self, message: Message) -> None:
        timings = [(f"queue.consumed by {self._app_id}", time.time())]

        logger.debug(
            "[MID:%s] Got `%s` message from `%s` service",
            message.message_id,
            message.type or "[untyped]",
            message.app_id or "unknown",
        )

        if message.type != "rpc.request":
            logger.warning(
                "[MID:%s] Unexpected message type: `%s`",
                message.message_id,
                message.type,
            )
            await message.reject()
            return

        if message.reply_to and not message.correlation_id:
            logger.warning("[MID:%s] `correlation_id` is not set", message.message_id)
            await message.reject()
            return

        published_timestamp = None
        if message.timestamp:
            published_timestamp = message.timestamp.timestamp()
            timings.insert(0, ("queue.published", published_timestamp))

        deadline = None
        if published_timestamp and message.expiration:
            deadline = published_timestamp + message.expiration

        try:
            timings.append(("execution.started", time.time()))
            response = await asyncio.wait_for(
                self._process_message(message),
                timeout=(deadline - time.time()) if deadline else None,
            )
            timings.append(("execution.completed", time.time()))
        except asyncio.TimeoutError:
            logger.info(
                "[MID:%s] [CID:%s] Execution of `%s` from `%s` service timed out",  # noqa: E501
                message.message_id,
                message.correlation_id,
                message.type,
                message.app_id or "unknown",
            )

            await message.ack()
            return

        response.timings += timings

        if message.reply_to:
            response_message = Message(
                body=response,
                app_id=self._app_id,
                priority=message.priority,
                correlation_id=message.correlation_id,
                expiration=(deadline - time.time()) if deadline else None,
                type="rpc.response",
            )

            await message.channel.basic_publish(
                body=response_message.body,
                routing_key=message.reply_to,
                properties=response_message.properties,
            )

        await message.ack()

        logger.info(
            "[MID:%s] [CID:%s] Handled `%s` from `%s` service, success: %s",  # noqa: E501
            message.message_id,
            message.correlation_id,
            message.type,
            message.app_id or "unknown",
            response.error is None,
        )

    @Response.wrap_errors
    async def _process_message(self, message: Message) -> Response:
        if message.content_type != "application/msgpack":
            raise errors.ParsingError(
                f"Got a message with unknown content type: {message.content_type}"
            )

        try:
            request = Request.model_validate_msgpack(message.body)
        except ValueError as error:
            raise errors.ParsingError(f"Message deserialization failed `{error!r}`")

        response = await self._process_request(request)
        response.context = request.context
        response.timings = request.timings

        logger.debug(
            "[MID:%s] Processed `%s` to method `%s`, success: %s",
            message.message_id,
            message.type,
            request.method_name,
            response.error is None,
        )

        return response

    @Response.wrap_errors
    async def _process_request(self, request) -> Response:
        handler = self._handlers.get(request.method_name)
        if handler is None:
            raise errors.UnknownMethodError(request.method_name)

        try:
            result = await handler(**request.arguments)
        except Exception as error:
            raise errors.ExecutionError(repr(error))

        if isinstance(result, pydantic.BaseModel):
            result = result.model_dump()

        return Response(result=result)

    async def _ping(self) -> None:
        return None

    async def _help(self):
        """
        Returns methods list
        """
        return {
            "methods": [
                {
                    "method_name": method_name,
                    "signature": f"{method_name}{inspect.signature(handler)}",
                    "description": handler.__doc__,
                }
                for method_name, handler in self._handlers.items()
            ]
        }



import functools
import logging
import msgpack
import pydantic

from typing import Dict, Any, List, Optional, Tuple

from . import errors


logger = logging.getLogger(__name__)


class BaseModel(pydantic.BaseModel):
    @classmethod
    def model_validate_msgpack(cls, obj: bytes):
        return cls.model_validate(msgpack.loads(obj))

    def model_dump_msgpack(self, **kwargs) -> bytes:
        kwargs.setdefault("exclude_defaults", True)
        return msgpack.dumps(self.model_dump(**kwargs))


class AbstractOptions(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", frozen=True)

    def __init__(self, *args: "AbstractOptions", **kwargs: Dict[str, Any]):
        if args:
            merged = {}
            self_class = type(self)
            for options in args:
                if not isinstance(options, self_class):
                    raise TypeError(
                        f"Positional arguments must be {self_class} instances"
                    )

                merged.update(options.model_dump(exclude_unset=True))
            merged.update(**kwargs)

            kwargs = merged

        return super().__init__(**kwargs)

    def update(self, *args: "AbstractOptions", **kwargs: Dict[str, Any]):
        if not args and not kwargs:
            return self

        return self.__class__(self, *args, **kwargs)


class Request(BaseModel):
    method_name: str
    arguments: Dict[str, Any] = {}
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []


class ResponseError(BaseModel):
    type: str
    message: Optional[str] = None


class Response(BaseModel):
    result: Optional[Any] = None
    error: Optional[ResponseError] = None
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []

    @classmethod
    def wrap_errors(cls, func):
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            try:
                response = await func(*args, **kwargs)
                if not isinstance(response, cls):
                    raise TypeError(f"Incorrect response type, got: {type(response)}")

            except errors.RpcError as error:
                response = cls(
                    error={"type": type(error).__name__, "message": error.args[0]}
                )

            except Exception as error:
                logger.exception("Unhandled error")
                response = cls(error={"type": "UnhandledError", "message": repr(error)})

            return response

        return inner



import aio_pika
import aio_pika.exceptions
import asyncio
import logging
import socket
import time
import uuid
import weakref

from typing import Any, Dict, Optional

from . import errors
from .models import AbstractOptions, Request, Response
from .types import Message
from .utils import fix_message_timestamp


logger = logging.getLogger(__name__)


class Options(AbstractOptions):
    timeout: Optional[float] = 60
    priority: Optional[int] = None


class RpcClient:
    def __init__(
        self, service_name, connection_string, *args: Options, **kwargs: Dict[str, Any]
    ):
        self._service_name = service_name
        self._connection_string = connection_string
        self._app_id = f"{service_name}@{socket.gethostname()}"

        self._options = Options(*args, **kwargs)
        self._connect_lock = asyncio.Lock()
        self._declare_exchange_lock = asyncio.Lock()

        self._reset()

    def _reset(self):
        self._connection = None
        self._channel = None
        self._reply_queue = None

        self._exchanges = {}
        self._futures = weakref.WeakValueDictionary()

    def get_service(self, service_name: str, *args: Options, **kwargs: Dict[str, Any]):
        options = self._options.update(*args, **kwargs)
        return RpcService(self, service_name, options)

    __getitem__ = __getattr__ = get_service

    async def _ensure_connection_and_queue(self):
        async with self._connect_lock:
            if self._connection:
                return

            connection = await aio_pika.connect_robust(self._connection_string)
            channel = await connection.channel(on_return_raises=True)

            queue_name = f"client.{self._service_name}.{uuid.uuid4().hex}"
            queue = await channel.declare_queue(
                name=queue_name,
                exclusive=True,
                auto_delete=True,
            )
            await queue.consume(fix_message_timestamp(self._on_message))

            self._connection = connection
            self._channel = channel
            self._reply_queue = queue

    async def stop(self):
        connection = self._connection
        if connection:
            self._reset()
            await connection.close()

    async def call(
        self,
        service_name: str,
        method_name: str,
        arguments: Optional[Dict[str, Any]] = None,
        options: Optional[Options] = None,
    ):
        if options.timeout:
            deadline = time.monotonic() + options.timeout
            waiter = lambda coro: asyncio.wait_for(  # noqa: E731
                coro, timeout=(deadline - time.monotonic())
            )
        else:
            waiter = lambda coro: coro  # noqa: E731

        correlation_id = str(uuid.uuid1())

        request = Request(method_name=method_name, arguments=arguments)

        future = asyncio.get_running_loop().create_future()
        self._futures[correlation_id] = future

        await waiter(self._send_request(correlation_id, service_name, request, options))

        try:
            return await waiter(future)
        except asyncio.CancelledError:
            logger.warning(
                "Wait was cancelled but not the request itself. "
                "Pass `timeout` option instead of using `asyncio.wait_for` or similar"
            )
            raise

    async def _send_request(self, correlation_id, service_name, request, options):
        await self._ensure_connection_and_queue()

        request_message = Message(
            body=request,
            app_id=self._app_id,
            priority=options.priority,
            correlation_id=correlation_id,
            reply_to=self._reply_queue.name,
            expiration=options.timeout or None,
            type="rpc.request",
        )

        async with self._declare_exchange_lock:
            if service_name not in self._exchanges:
                try:
                    self._exchanges[service_name] = await self._channel.get_exchange(
                        f"service.{service_name}"
                    )
                except aio_pika.exceptions.ChannelNotFoundEntity:
                    raise errors.UnknownServiceError("Exchange was not found")

        try:
            await self._exchanges[service_name].publish(
                message=request_message,
                routing_key=request.method_name,
            )
        except aio_pika.exceptions.PublishError:
            raise errors.UnknownServiceError("Request was not delivered to a queue")

    async def _on_message(self, message):
        if message.type != "rpc.response":
            logger.warning(
                "[MID:%s] Unexpected message type: `%s`",
                message.message_id,
                message.type,
            )
            await message.reject()
            return

        if not message.correlation_id:
            logger.warning("[MID:%s] `correlation_id` is not set", message.message_id)
            await message.reject()
            return

        future = self._futures.pop(message.correlation_id, None)
        if not future or future.done():
            logger.warning(
                "[MID:%s] [CID:%s] Response is not awaited (too late or duplicate)",
                message.message_id,
                message.correlation_id,
            )
            await message.reject()
            return

        try:
            future.set_result(await self._process_message(message))
        except Exception as error:
            future.set_exception(error)

        await message.ack()

        logger.debug(
            "[MID:%s] [CID:%s] Got `%s` from `%s` service",
            message.message_id,
            message.correlation_id,
            message.type,
            message.app_id or "unknown",
        )

    async def _process_message(self, message) -> Any:
        if message.content_type != "application/msgpack":
            raise errors.ParsingError(
                f"Got a message with unknown content type: {message.content_type}"
            )

        try:
            response = Response.model_validate_msgpack(message.body)
        except ValueError as error:
            raise errors.ParsingError(f"Message deserialization failed `{error!r}`")

        if response.error:
            error_class = getattr(errors, response.error.type, errors.UnknownError)
            raise error_class(response.error.message)

        return response.result


class RpcService:
    def __init__(self, client, service_name, options):
        self._client = client
        self._service_name = service_name
        self._options = options

    def get_method(self, method_name: str, *args: Options, **kwargs: Dict[str, Any]):
        options = self._options.update(*args, **kwargs)
        return RpcMethod(self._client, self._service_name, method_name, options)

    __getitem__ = __getattr__ = get_method

    def __repr__(self):
        class_ = self.__class__
        return f"<{class_.__qualname__} `{self._service_name}`>"


class RpcMethod:
    def __init__(self, client, service_name, method_name, options):
        self._client = client
        self._service_name = service_name
        self._method_name = method_name
        self._options = options

    def __repr__(self):
        class_ = self.__class__
        return (
            f"<{class_.__qualname__} `{self._method_name}` "
            "of service `{self._service_name}`>"
        )

    def __call__(self, *args: Options, **kwargs: Dict[str, Any]):
        options = self._options.update(*args)
        return self._client.call(
            service_name=self._service_name,
            method_name=self._method_name,
            arguments=kwargs,
            options=options,
        )



import datetime
import functools


def fix_message_timestamp(func):
    @functools.wraps(func)
    def inner(message):
        # Fix `pamqp` naive timestamp
        if message.timestamp:
            message.timestamp = message.timestamp.replace(tzinfo=datetime.timezone.utc)

        return func(message)

    return inner
